# -*- coding: utf-8 -*-

__version__ = '2.1.0'
program_name = 'bio_hansel'
program_summary = 'Subtype microbial genomes using SNV targeting k-mer subtyping schemes.'
program_desc = program_summary + '''
Includes schemes for Salmonella enterica spp. enterica serovar Heidelberg and Enteritidis subtyping.
Developed by Geneviève Labbé, James Robertson, Peter Kruczkiewicz, Marisa Rankin, Matthew Gopez, Chad R. Laing, Philip Mabon, Kim Ziebell, Aleisha R. Reimer, Lorelee Tschetter, Gary Van Domselaar, Sadjia Bekal, Kimberley A. MacDonald, Linda Hoang, Linda Chui, Danielle Daignault, Durda Slavic, Frank Pollari, E. Jane Parmley, Philip Mabon, Elissa Giang, Lok Kan Lee, Jonathan Moffat, Marisa Rankin, Joanne MacKinnon, Roger Johnson, John H.E. Nash.
'''



# -*- coding: utf-8 -*-
import logging
from typing import Dict, List, Set

import attr
from collections import defaultdict

from .parsers import parse_fasta


@attr.s
class SubtypeCounts:
    subtype = attr.ib()
    refpositions = attr.ib(default=None)
    subtype_tile_count = attr.ib(default=0, validator=attr.validators.instance_of(int))
    positive_tile_count = attr.ib(default=0, validator=attr.validators.instance_of(int))
    negative_tile_count = attr.ib(default=0, validator=attr.validators.instance_of(int))
    all_tile_count = attr.ib(default=0, validator=attr.validators.instance_of(int))

    @subtype.validator
    def _check_subtype(self, attribute, value):
        if value is None or value == '':
            raise ValueError('Subtype cannot be None or empty string')
        if len(value) > 1:
            if '.' not in value:
                raise ValueError(
                    'Invalid subtype specified! "{}" does not numbers delimited by "." (periods)'.format(value))

    @subtype_tile_count.validator
    def _check_subtype_tile_count(self, attribute, value):
        if value == 0:
            raise ValueError('Subtype tile count cannot be zero!')

    @positive_tile_count.validator
    def _check_positive_tile_count(self, attribute, value):
        if value < self.subtype_tile_count:
            raise ValueError(
                'Subtype {}: Number of all subtype positive tiles (n={}) cannot be less than the number of subtype specific tiles (n={})'.format(
                    self.subtype,
                    value,
                    self.subtype_tile_count))
        if value > self.all_tile_count:
            raise ValueError(
                'Subtype {}: Number of all subtype positive tiles (n={}) cannot exceed number of all matching tiles for subtype (n={})'.format(
                    self.subtype,
                    value,
                    self.all_tile_count))
        if value == 0:
            raise ValueError('Subtype {}: Number of all subtype positive tiles cannot be zero!'.format(self.subtype))


def _tiles(tiles_fasta: str) -> (Dict[str, List[str]], Dict[str, List[str]], Set[int]):
    tiles = defaultdict(list)
    neg_tiles = defaultdict(list)
    sizes = set()
    for h, s in parse_fasta(tiles_fasta):
        sizes.add(len(s))
        _, st = h.split('-')
        if 'negative' not in h:
            tiles[st].append(h)
        else:
            neg_tiles[st].append(h)
    return tiles, neg_tiles, sizes


def subtype_counts(scheme_fasta: str) -> Dict[str, SubtypeCounts]:
    subtype_counts = {}
    tiles, neg_tiles, sizes = _tiles(scheme_fasta)
    if len(sizes) > 1:
        logging.warning('Not all markers in "%s" of the same size! %s', scheme_fasta, sizes)
    n_tiles_total = sum([len(vs) for vs in tiles.values()])
    ks = [x for x in tiles.keys()]
    ks.sort()
    for k in ks:
        st_pos_count_rest = 0
        subtypes_set = {k}
        if len(k) > 1:
            *r, _ = k.split('.')
            for i in range(len(r)):
                sub_k = '.'.join(r[:i + 1])
                subtypes_set.add(sub_k)
                pos_tiles = tiles[sub_k]
                st_pos_count_rest += len(pos_tiles)
        st_count = len(tiles[k])
        st_count_pos = st_pos_count_rest + st_count
        st_neg_count = sum([len(v) for k,v in neg_tiles.items() if k not in subtypes_set])

        subtype_count = SubtypeCounts(subtype=k,
                                      refpositions={int(v.split('-')[0]) for v in tiles[k]},
                                      subtype_tile_count=st_count,
                                      positive_tile_count=st_count_pos,
                                      negative_tile_count=st_neg_count,
                                      all_tile_count=st_neg_count + st_count_pos)
        subtype_counts[k] = subtype_count
    return subtype_counts



import logging
from typing import Optional

import numpy as np
import os
import pandas as pd


def read_metadata_table(path: str) -> Optional[pd.DataFrame]:
    """Read a subtype metadata table into a Pandas DataFrame.

    This table must have a column labeled `subtype` and have one of the following file extensions:

    - `.tab` - tab-delimited file
    - `.tsv` - tab-delimited file
    - `.csv` - comma-separated values file

    The top row must be the header row.

    Args:
        path: File path of table.

    Returns:
        DataFrame of table file if `path` is one of the acceptable file formats, otherwise, return `None`
    """
    FILE_EXT_TO_PD_READ_FUNC = {
        '.tab': pd.read_table,
        '.tsv': pd.read_table,
        '.csv': pd.read_csv
    }
    _, file_ext = os.path.splitext(os.path.basename(path))
    file_ext = file_ext.lower()
    if file_ext not in FILE_EXT_TO_PD_READ_FUNC:
        logging.error('File extension of metadata file "{}" not one of the expected "{}"'.format(
            path,
            list(FILE_EXT_TO_PD_READ_FUNC.keys())
        ))
        return None
    dfmd = FILE_EXT_TO_PD_READ_FUNC[file_ext](path)  # type: pd.DataFrame
    assert np.any(dfmd.columns == 'subtype'), 'Column with name "subtype" expected in metadata file "{}"'.format(path)
    dfmd.subtype = dfmd.subtype.astype(str)
    logging.info('Read scheme metadata file "{}" into DataFrame with shape {}'.format(path, dfmd.shape))
    return dfmd


def merge_metadata_with_summary_results(df_results: pd.DataFrame, df_metadata: pd.DataFrame) -> pd.DataFrame:
    """Merge subtype metadata table into subtype results table.

    Args:
        df_results: Subtyping results table.
        df_metadata: Subtype metadata table.

    Returns:
        Subtyping results with subtype metadata merged in if metadata is present for subtype results.
    """
    df_results.subtype = df_results.subtype.fillna('')
    df_results.subtype = df_results.subtype.astype(str)
    return pd.merge(df_results, df_metadata, how='left', on='subtype')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import logging
import sys
import re
import os
from typing import Optional, List, Any, Tuple

import attr
import pandas as pd

from . import program_desc, __version__
from .const import SUBTYPE_SUMMARY_COLS, REGEX_FASTQ, REGEX_FASTA, JSON_EXT_TMPL
from .subtype import Subtype
from .subtype_stats import subtype_counts
from .subtyper import \
    subtype_contigs_samples, \
    subtype_reads_samples
from .metadata import read_metadata_table, merge_metadata_with_summary_results
from .utils import \
    genome_name_from_fasta_path, \
    get_scheme_fasta, \
    does_file_exist, \
    collect_fastq_from_dir, \
    group_fastqs, \
    collect_fasta_from_dir, \
    init_subtyping_params

SCRIPT_NAME = 'hansel'
LOG_FORMAT = '%(asctime)s %(levelname)s: %(message)s [in %(pathname)s:%(lineno)d]'


def init_console_logger(logging_verbosity=3):
    logging_levels = [logging.ERROR, logging.WARN, logging.INFO, logging.DEBUG]
    if logging_verbosity > (len(logging_levels) - 1):
        logging_verbosity = 3
    lvl = logging_levels[logging_verbosity]

    logging.basicConfig(format=LOG_FORMAT, level=lvl)


def init_parser():
    parser = argparse.ArgumentParser(prog=SCRIPT_NAME,
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description=program_desc)
    parser.add_argument('files',
                        metavar='F',
                        nargs='*',
                        help='Input genome FASTA/FASTQ files (can be Gzipped)')
    parser.add_argument('-s', '--scheme',
                        default='heidelberg',
                        help='Scheme to use for subtyping (built-in: "heidelberg", "enteritidis"; OR user-specified: '
                             '/path/to/user/scheme)')
    parser.add_argument('--scheme-name',
                        help='Custom user-specified SNP substyping scheme name')
    parser.add_argument('-M', '--scheme-metadata',
                        help='Scheme subtype metadata table (CSV or tab-delimited format; must contain "subtype" column)')
    parser.add_argument('-p', '--paired-reads',
                        nargs=2,
                        metavar=('forward_reads', 'reverse_reads'),
                        action='append',
                        help='FASTQ paired-end reads')
    parser.add_argument('-i', '--input-fasta-genome-name',
                        nargs=2,
                        metavar=('fasta_path', 'genome_name'),
                        action='append',
                        help='fasta file path to genome name pair')
    parser.add_argument('-D', '--input-directory',
                        help='directory of input fasta files (.fasta|.fa|.fna) or FASTQ files (paired FASTQ should '
                             'have same basename with "_\d\.(fastq|fq)" postfix to be automatically paired) '
                             '(files can be Gzipped)')
    parser.add_argument('-o', '--output-summary',
                        help='Subtyping summary output path (tab-delimited)')
    parser.add_argument('-O', '--output-tile-results',
                        help='Subtyping tile matching output path (tab-delimited)')
    parser.add_argument('-S', '--output-simple-summary',
                        help='Subtyping simple summary output path')
    parser.add_argument('--force',
                        action='store_true',
                        help='Force existing output files to be overwritten')
    parser.add_argument('--json',
                        action='store_true',
                        help='Output JSON representation of output files')
    parser.add_argument('--min-kmer-freq',
                        type=int,
                        help='Min k-mer freq/coverage')
    parser.add_argument('--max-kmer-freq',
                        type=int,
                        help='Max k-mer freq/coverage')
    parser.add_argument('--low-cov-depth-freq',
                        type=int,
                        help='Frequencies below this coverage are considered low coverage')
    parser.add_argument('--max-missing-tiles',
                        type=float,
                        help='Decimal proportion of maximum allowable missing tiles before being considered an error. (0.0 - 1.0)')
    parser.add_argument('--min-ambiguous-tiles',
                        type=int,
                        help='Minimum number of missing tiles to be considered an ambiguous result')
    parser.add_argument('--low-cov-warning',
                        type=int,
                        help='Overall tile coverage below this value will trigger a low coverage warning')
    parser.add_argument('--max-intermediate-tiles',
                        type=float,
                        help='Decimal proportion of maximum allowable missing tiles to be considered an intermediate subtype. (0.0 - 1.0)')
    parser.add_argument('-t', '--threads',
                        type=int,
                        default=1,
                        help='Number of parallel threads to run analysis (default=1)')
    parser.add_argument('-v', '--verbose',
                        action='count',
                        default=0,
                        help='Logging verbosity level (-v == show warnings; -vvv == show debug info)')
    parser.add_argument('-V', '--version',
                        action='version',
                        version='%(prog)s {}'.format(__version__))
    return parser


def collect_inputs(args: Any) -> Tuple[List[Tuple[str, str]], List[Tuple[List[str], str]]]:
    """Collect all input files for analysis

    Sample names are derived from the base filename with no extensions.
    Sequencing reads are paired if they share a common filename name without "_\d".
    Filepaths for contigs and reads files are collected from an input directory if provided.

    Args:
        args: ArgumentParser.parse_args() output

    Returns:
        List of (contig filename, sample name)
        List of ([reads filepaths], sample name)
    """
    input_genomes = []
    reads = []
    if args.files:
        fastas = [x for x in args.files if REGEX_FASTA.match(x)]
        fastqs = [x for x in args.files if REGEX_FASTQ.match(x)]
        if len(fastas) > 0:
            logging.info('# of input fastas %s', len(fastas))
            for fasta_path in fastas:
                fasta_path = os.path.abspath(fasta_path)
                if os.path.exists(fasta_path):
                    genome_name = genome_name_from_fasta_path(fasta_path)
                    input_genomes.append((fasta_path, genome_name))
                else:
                    logging.error('Input fasta "%s" does not exist!', fasta_path)
        if len(fastqs) > 0:
            logging.info('# of input fastqs %s', len(fastqs))
            grouped_fastqs = group_fastqs(fastqs)
            logging.info('Grouped %s fastqs into %s groups',
                         len(fastqs),
                         len(grouped_fastqs))
            reads += grouped_fastqs
    if args.input_fasta_genome_name:
        for fasta_path, genome_name in args.input_fasta_genome_name:
            input_genomes.append((os.path.abspath(fasta_path), genome_name))
    if args.input_directory:
        logging.info('Searching dir "%s" for FASTA files', args.input_directory)
        input_genomes += collect_fasta_from_dir(args.input_directory)
        logging.info('Searching dir "%s" for FASTQ files', args.input_directory)
        reads += collect_fastq_from_dir(args.input_directory)
    if args.paired_reads:
        for x in args.paired_reads:
            if not isinstance(x, (list, tuple)):
                logging.warning('Paired end reads not list or tuple %s', x)
                continue
            filenames = [os.path.basename(y) for y in x]
            common_prefix = os.path.commonprefix(filenames)
            genome_name = re.sub(r'[\W\_]+$', r'', common_prefix)
            if genome_name == '':
                genome_name = filenames[0]
            reads.append((x, genome_name))
    return input_genomes, reads


def main():
    parser = init_parser()
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()
    args = parser.parse_args()
    init_console_logger(args.verbose)
    output_summary_path = args.output_summary
    output_tile_results = args.output_tile_results
    output_simple_summary_path = args.output_simple_summary
    does_file_exist(output_simple_summary_path, args.force)
    does_file_exist(output_summary_path, args.force)
    does_file_exist(output_tile_results, args.force)
    scheme = args.scheme  # type: str
    scheme_name = args.scheme_name  # type: Optional[str]
    scheme_fasta = get_scheme_fasta(scheme)
    scheme_subtype_counts = subtype_counts(scheme_fasta)
    logging.debug(args)
    subtyping_params = init_subtyping_params(args, scheme)
    input_contigs, input_reads = collect_inputs(args)
    if len(input_contigs) == 0 and len(input_reads) == 0:
        raise Exception('No input files specified!')
    df_md = None
    if args.scheme_metadata:
        df_md = read_metadata_table(args.scheme_metadata)
    n_threads = args.threads

    subtype_results = []  # type: List[Tuple[Subtype, pd.DataFrame]]
    if len(input_contigs) > 0:
        contigs_results = subtype_contigs_samples(input_genomes=input_contigs,
                                                  scheme=scheme,
                                                  scheme_name=scheme_name,
                                                  subtyping_params=subtyping_params,
                                                  scheme_subtype_counts=scheme_subtype_counts,
                                                  n_threads=n_threads)
        logging.info('Generated %s subtyping results from %s contigs samples', len(contigs_results), len(input_contigs))
        subtype_results += contigs_results
    if len(input_reads) > 0:
        reads_results = subtype_reads_samples(reads=input_reads,
                                              scheme=scheme,
                                              scheme_name=scheme_name,
                                              subtyping_params=subtyping_params,
                                              scheme_subtype_counts=scheme_subtype_counts,
                                              n_threads=n_threads)
        logging.info('Generated %s subtyping results from %s contigs samples', len(reads_results), len(input_reads))
        subtype_results += reads_results

    dfs = [df for st, df in subtype_results]  # type: List[pd.DataFrame]
    dfsummary = pd.DataFrame([attr.asdict(st) for st, df in subtype_results])
    dfsummary = dfsummary[SUBTYPE_SUMMARY_COLS]

    if dfsummary['avg_tile_coverage'].isnull().all():
        dfsummary = dfsummary.drop(labels='avg_tile_coverage', axis=1)

    if df_md is not None:
        dfsummary = merge_metadata_with_summary_results(dfsummary, df_md)

    kwargs_for_pd_to_table = dict(sep='\t', index=None, float_format='%.3f')
    kwargs_for_pd_to_json = dict(orient='records')

    if output_summary_path:
        dfsummary.to_csv(output_summary_path, **kwargs_for_pd_to_table)
        if args.json:
            dfsummary.to_json(JSON_EXT_TMPL.format(output_summary_path), **kwargs_for_pd_to_json)
        logging.info('Wrote subtyping output summary to %s', output_summary_path)
    else:
        # if no output path specified for the summary results, then print to stdout
        print(dfsummary.to_csv(sep='\t', index=None))

    if output_tile_results:
        if len(dfs) > 0:
            dfall = pd.concat(dfs)  # type: pd.DataFrame
            dfall.to_csv(output_tile_results, **kwargs_for_pd_to_table)
            logging.info('Tile results written to "{}".'.format(output_tile_results))
            if args.json:
                dfall.to_json(JSON_EXT_TMPL.format(output_tile_results), **kwargs_for_pd_to_json)
                logging.info(
                    'Tile results written to "{}" in JSON format.'.format(JSON_EXT_TMPL.format(output_tile_results)))
        else:
            logging.error(
                'No tile results generated. No tile results file written to "{}".'.format(output_tile_results))

    if output_simple_summary_path:
        if 'avg_tile_coverage' in dfsummary.columns:
            df_simple_summary = dfsummary[['sample', 'subtype', 'avg_tile_coverage', 'qc_status', 'qc_message']]
        else:
            df_simple_summary = dfsummary[['sample', 'subtype', 'qc_status', 'qc_message']]

        if df_md is not None:
            df_simple_summary = merge_metadata_with_summary_results(df_simple_summary, df_md)

        df_simple_summary.to_csv(output_simple_summary_path, **kwargs_for_pd_to_table)
        if args.json:
            df_simple_summary.to_json(JSON_EXT_TMPL.format(output_simple_summary_path), **kwargs_for_pd_to_json)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

class QC:
    FAIL = 'FAIL'
    WARNING = 'WARNING'
    PASS = 'PASS'
    AMBIGUOUS_RESULTS_ERROR_3 = 'Ambiguous Results Error 3'
    UNCONFIDENT_RESULTS_ERROR_4 = 'Unconfident Results Error 4'
    NO_SUBTYPE_RESULT = 'No subtype result!'
    NO_TARGETS_FOUND = 'No tiles/targets were found in this sample.'



# -*- coding: utf-8 -*-
"""
Functions for subtyping of reads (e.g. FASTQ) and contigs (e.g. FASTA) using bio_hansel-compatible subtyping schemes.
"""
import logging
from typing import Optional, List, Dict, Union, Tuple

import pandas as pd
import re

from .aho_corasick import init_automaton, find_in_fasta, find_in_fastqs
from .const import COLUMNS_TO_REMOVE
from .qc import perform_quality_check, QC
from .subtype import Subtype
from .subtype_stats import SubtypeCounts
from .subtype_stats import subtype_counts
from .subtyping_params import SubtypingParams
from .utils import find_inconsistent_subtypes, get_scheme_fasta, get_scheme_version, init_subtyping_params


def subtype_reads_samples(reads: List[Tuple[List[str], str]],
                          scheme: str,
                          scheme_name: Optional[str] = None,
                          subtyping_params: Optional[SubtypingParams] = None,
                          scheme_subtype_counts: Optional[Dict[str, SubtypeCounts]] = None,
                          n_threads: int = 1) -> List[Tuple[Subtype, pd.DataFrame]]:
    """Subtype input genomes using a scheme.

    Args:
        reads: input genomes; tuple of list of FASTQ file paths and genome name
        scheme: bio_hansel scheme FASTA path
        scheme_name: optional scheme name
        subtyping_params: scheme specific subtyping parameters
        scheme_subtype_counts: summary information about scheme
        n_threads: number of threads to use for subtyping analysis

    Returns:
        List of tuple of Subtype and detailed subtyping results for each sample
    """
    if n_threads == 1:
        logging.info('Serial single threaded run mode on %s input genomes', len(reads))
        outputs = [subtype_reads(reads=fastq_files,
                                 genome_name=genome_name,
                                 scheme=scheme,
                                 scheme_name=scheme_name,
                                 subtyping_params=subtyping_params,
                                 scheme_subtype_counts=scheme_subtype_counts)
                   for fastq_files, genome_name in reads]
    else:
        outputs = parallel_query_reads(reads=reads,
                                       scheme=scheme,
                                       scheme_name=scheme_name,
                                       subtyping_params=subtyping_params,
                                       scheme_subtype_counts=scheme_subtype_counts,
                                       n_threads=n_threads)
    return outputs


def subtype_contigs_samples(input_genomes: List[Tuple[str, str]],
                            scheme: str,
                            scheme_name: Optional[str] = None,
                            subtyping_params: Optional[SubtypingParams] = None,
                            scheme_subtype_counts: Optional[Dict[str, SubtypeCounts]] = None,
                            n_threads: int = 1) -> List[Tuple[Subtype, pd.DataFrame]]:
    """Subtype input genomes using a scheme.

    Args:
        input_genomes: input genomes; tuple of FASTA file path and genome name
        scheme: bio_hansel scheme FASTA path
        scheme_name: optional scheme name
        subtyping_params: scheme specific subtyping parameters
        scheme_subtype_counts: summary information about scheme
        n_threads: number of threads to use for subtyping analysis

    Returns:
        List of tuple of Subtype and detailed subtyping results for each sample
    """
    if n_threads == 1:
        logging.info('Serial single threaded run mode on %s input genomes', len(input_genomes))
        outputs = [subtype_contigs(fasta_path=input_fasta,
                                   genome_name=genome_name,
                                   scheme=scheme,
                                   scheme_name=scheme_name,
                                   subtyping_params=subtyping_params,
                                   scheme_subtype_counts=scheme_subtype_counts)
                   for input_fasta, genome_name in input_genomes]
    else:
        outputs = parallel_query_contigs(input_genomes, scheme, scheme_name, subtyping_params, scheme_subtype_counts,
                                         n_threads)
    return outputs


def subtype_contigs(fasta_path: str,
                    genome_name: str,
                    scheme: str,
                    subtyping_params: Optional[SubtypingParams] = None,
                    scheme_name: Optional[str] = None,
                    scheme_subtype_counts: Optional[Dict[str, SubtypeCounts]] = None) -> Tuple[
    Subtype, pd.DataFrame]:
    """Subtype input contigs using a particular scheme.

    Args:
        fasta_path: Input FASTA file path
        genome_name: Input genome name
        scheme: bio_hansel scheme FASTA path
        subtyping_params: scheme specific subtyping parameters
        scheme_name: optional scheme name
        scheme_subtype_counts: summary information about scheme

    Returns:
        - Subtype result
        - pd.DataFrame of detailed subtyping results
    """
    scheme_fasta = get_scheme_fasta(scheme)
    if scheme_subtype_counts is None:
        scheme_subtype_counts = subtype_counts(scheme_fasta)
    if subtyping_params is None:
        subtyping_params = init_subtyping_params(scheme=scheme)
    scheme_version = get_scheme_version(scheme)
    st = Subtype(sample=genome_name,
                 file_path=fasta_path,
                 scheme=scheme_name or scheme,
                 scheme_version=scheme_version,
                 scheme_subtype_counts=scheme_subtype_counts)

    automaton = init_automaton(scheme_fasta)
    df = find_in_fasta(automaton, fasta_path)

    if df is None or df.shape[0] == 0:
        logging.warning('No subtyping tile matches for input "%s" for scheme "%s"', fasta_path, scheme)
        st.qc_status = QC.FAIL
        st.qc_message = QC.NO_TARGETS_FOUND
        st.are_subtypes_consistent = False
        return st, empty_results(st)

    refpositions = [x for x, y in df.tilename.str.split('-')]
    subtypes = [y for x, y in df.tilename.str.split('-')]
    df['refposition'] = [int(x.replace('negative', '')) for x in refpositions]
    df['subtype'] = subtypes
    df['is_pos_tile'] = ~df.tilename.str.contains('negative')
    process_subtyping_results(st, df, scheme_subtype_counts)
    st.qc_status, st.qc_message = perform_quality_check(st, df, subtyping_params)

    logging.info(st)

    df['sample'] = genome_name
    df['file_path'] = fasta_path
    df['scheme'] = scheme_name or scheme
    df['scheme_version'] = scheme_version
    df['qc_status'] = st.qc_status
    df['qc_message'] = st.qc_message

    df = df[df.columns[~df.columns.isin(COLUMNS_TO_REMOVE)]]
    return st, df


def parallel_query_contigs(input_genomes: List[Tuple[str, str]],
                           scheme: str,
                           scheme_name: Optional[str] = None,
                           subtyping_params: Optional[SubtypingParams] = None,
                           scheme_subtype_counts: Optional[Dict[str, SubtypeCounts]] = None,
                           n_threads: int = 1):
    """Parallel subtyping of input contigs

    Subtype and analyse each input in parallel using a multiprocessing thread pool.

    Args:
        input_genomes: Input genome FASTA paths
        scheme: bio_hansel scheme FASTA path to subtype with
        scheme_name: optional scheme name
        subtyping_params: scheme specific subtyping parameters
        scheme_subtype_counts: scheme summary information
        n_threads: Number of threads to use

    Returns:
        A list of tuples of Subtype results and a pd.DataFrame of detailed subtyping results for each input
    """
    from multiprocessing import Pool
    logging.info('Initializing thread pool with %s threads', n_threads)
    pool = Pool(processes=n_threads)
    logging.info('Running analysis asynchronously on %s input genomes', len(input_genomes))
    res = [pool.apply_async(subtype_contigs, (input_fasta,
                                              genome_name,
                                              scheme,
                                              subtyping_params,
                                              scheme_name,
                                              scheme_subtype_counts))
           for input_fasta, genome_name in input_genomes]
    logging.info('Parallel analysis complete! Retrieving analysis results')
    outputs = [x.get() for x in res]
    return outputs


def parallel_query_reads(reads: List[Tuple[List[str], str]],
                         scheme: str,
                         scheme_name: Optional[str] = None,
                         subtyping_params: Optional[SubtypingParams] = None,
                         scheme_subtype_counts: Optional[Dict[str, SubtypeCounts]] = None,
                         n_threads: int = 1) -> List[Tuple[Subtype, pd.DataFrame]]:
    """Parallel subtyping of input reads

    Subtype and analyse each input in parallel using a multiprocessing thread pool.

    Args:
        reads: Input reads; list of tuples of FASTQ file paths and genome names
        scheme: bio_hansel scheme FASTA path
        scheme_name: optional scheme name
        subtyping_params: scheme specific subtyping parameters
        scheme_subtype_counts: scheme summary information
        n_threads: number of threads to use

    Returns:
        A list of tuples of Subtype results and a pd.DataFrame of detailed subtyping results for each input
    """
    from multiprocessing import Pool
    logging.info('Initializing thread pool with %s threads', n_threads)
    pool = Pool(processes=n_threads)
    logging.info('Running analysis asynchronously on %s input genomes', len(reads))
    res = [pool.apply_async(subtype_reads, (fastqs,
                                            genome_name,
                                            scheme,
                                            scheme_name,
                                            subtyping_params,
                                            scheme_subtype_counts))
           for fastqs, genome_name in reads]
    logging.info('Parallel analysis complete! Retrieving analysis results')
    outputs = [x.get() for x in res]
    return outputs


def subtype_reads(reads: Union[str, List[str]],
                  genome_name: str,
                  scheme: str,
                  scheme_name: Optional[str] = None,
                  subtyping_params: Optional[SubtypingParams] = None,
                  scheme_subtype_counts: Optional[Dict[str, SubtypeCounts]] = None) -> Tuple[
    Subtype, Optional[pd.DataFrame]]:
    """Subtype input reads using a particular scheme.

    Args:
        reads: Input FASTQ file path(s)
        genome_name: Input genome name
        scheme: bio_hansel scheme FASTA path
        scheme_name: optional scheme name
        subtyping_params: scheme specific subtyping parameters
        scheme_subtype_counts: summary information about scheme

    Returns:
        - Subtype result
        - pd.DataFrame of detailed subtyping results
    """
    scheme_fasta = get_scheme_fasta(scheme)
    if scheme_subtype_counts is None:
        scheme_subtype_counts = subtype_counts(scheme_fasta)
    if subtyping_params is None:
        subtyping_params = init_subtyping_params(scheme=scheme)
    scheme_version = get_scheme_version(scheme)

    st = Subtype(sample=genome_name,
                 file_path=reads,
                 scheme=scheme_name or scheme,
                 scheme_version=scheme_version,
                 scheme_subtype_counts=scheme_subtype_counts)

    automaton = init_automaton(scheme_fasta)
    if isinstance(reads, str):
        df = find_in_fastqs(automaton, reads)
    elif isinstance(reads, list):
        df = find_in_fastqs(automaton, *reads)
    else:
        raise ValueError('Unexpected type "{}" for "reads": {}'.format(type(reads), reads))

    if df is None or df.shape[0] == 0:
        logging.warning('No subtyping tile matches for input "%s" for scheme "%s"', reads, scheme)
        st.are_subtypes_consistent = False
        st.qc_status = QC.FAIL
        st.qc_message = QC.NO_TARGETS_FOUND
        return st, empty_results(st)

    refpositions = [x for x, y in df.tilename.str.split('-')]
    subtypes = [y for x, y in df.tilename.str.split('-')]
    df['refposition'] = [int(x.replace('negative', '')) for x in refpositions]
    df['subtype'] = subtypes
    df['is_pos_tile'] = ~df.tilename.str.contains('negative')
    df['is_kmer_freq_okay'] = (df.freq >= subtyping_params.min_kmer_freq) & (df.freq <= subtyping_params.max_kmer_freq)
    st.avg_tile_coverage = df['freq'].mean()
    st, df = process_subtyping_results(st, df[df.is_kmer_freq_okay], scheme_subtype_counts)
    st.qc_status, st.qc_message = perform_quality_check(st, df, subtyping_params)
    df['file_path'] = str(st.file_path)
    df['sample'] = genome_name
    df['scheme'] = scheme_name or scheme
    df['scheme_version'] = scheme_version
    df['qc_status'] = st.qc_status
    df['qc_message'] = st.qc_message
    df = df[df.columns[~df.columns.isin(COLUMNS_TO_REMOVE)]]
    return st, df


def process_subtyping_results(st: Subtype, df: pd.DataFrame, scheme_subtype_counts: Dict[str, SubtypeCounts]) -> Tuple[
    Subtype, pd.DataFrame]:
    """Process the subtyping results to get the final subtype result and summary stats

    Args:
        st: Subtype result
        df: Subtyping results
        scheme_subtype_counts: Subtyping scheme summary info
    Returns:
        Tuple of `st` and `df`
    """
    dfpos = df[df.is_pos_tile]
    dfpos_highest_res = highest_resolution_subtype_results(dfpos)
    subtype_list = [x for x in dfpos_highest_res.subtype.unique()]
    st = set_subtype_results(st, dfpos, subtype_list)
    st = set_inconsistent_subtypes(st, find_inconsistent_subtypes(sorted_subtype_ints(dfpos.subtype)))
    st = set_subtyping_stats(st, df, dfpos, dfpos_highest_res, subtype_list, scheme_subtype_counts)
    st.non_present_subtypes = absent_downstream_subtypes(st.subtype, df.subtype, list(scheme_subtype_counts.keys()))
    return st, df


def set_subtype_results(st: Subtype, df_positive: pd.DataFrame, subtype_list: List[str]) -> Subtype:
    """Set subtype results

    Args:
        st: Subtype result
        df_positive: Positive tiles subtyping results
        subtype_list: List of subtypes found
    """
    st.subtype = '; '.join(subtype_list)
    st.tiles_matching_subtype = '; '.join(subtype_list)
    pos_subtypes_str = [x for x in df_positive.subtype.unique()]
    pos_subtypes_str.sort(key=lambda x: len(x))
    st.all_subtypes = '; '.join(pos_subtypes_str)
    return st


def set_subtyping_stats(st: Subtype,
                        df: pd.DataFrame,
                        dfpos: pd.DataFrame,
                        dfpos_highest_res: pd.DataFrame,
                        subtype_list: List[str],
                        scheme_subtype_counts: Dict[str, SubtypeCounts]) -> Subtype:
    """Set subtyping result stats

    - `n_tiles_matching_subtype`: # tiles matching final subtype
    - `n_tiles_matching_all`: # tiles found
    - `n_tiles_matching_positive`: # positive tiles found
    - `n_tiles_matching_negative`:  # negative tiles found
    - `n_tiles_matching_all_expected`: expected # of all tiles found for each subtype
    - `n_tiles_matching_positive_expected`: expected # of positive tiles found for each subtype
    - `n_tiles_matching_subtype_expected`: expected # of subtype-specific tiles found for each subtype

    Args:
        st: Subtype result
        df: Subtyping results
        dfpos: Positive tile subtyping results
        dfpos_highest_res: Final subtype specific subtyping results
        subtype_list: List of subtypes found
        scheme_subtype_counts: Subtyping scheme summary info
    """
    st.n_tiles_matching_subtype = dfpos_highest_res.shape[0]
    st.n_tiles_matching_all = df.tilename.unique().size
    st.n_tiles_matching_positive = dfpos.tilename.unique().size
    st.n_tiles_matching_negative = df[~df.is_pos_tile].shape[0]
    st.n_tiles_matching_all_expected = ';'.join([str(scheme_subtype_counts[x].all_tile_count) for x in subtype_list])
    st.n_tiles_matching_positive_expected = ';'.join(
        [str(scheme_subtype_counts[x].positive_tile_count) for x in subtype_list])
    st.n_tiles_matching_subtype_expected = ';'.join(
        [str(scheme_subtype_counts[x].subtype_tile_count) for x in subtype_list])
    return st


def count_periods(s: str) -> int:
    """Count the number of periods in a string.

    Examples:

        count_periods("2.1.1") => 2
        count_periods("1") => 0

    Args:
        s: Some string with periods

    Returns:
        The number of periods found in the input string.
    """
    return sum((1 for c in list(s) if c == '.'))


def highest_resolution_subtype_results(df: pd.DataFrame) -> pd.DataFrame:
    """Get the highest resolution subtype results

    Where the highest resolution result has the most periods ('.') in its designation.

    Args:
        df: positive subtyping results

    Returns:
        Highest resolution (most periods in subtype) subtyping results
    """
    subtype_lens = df.subtype.apply(count_periods)
    max_subtype_strlen = subtype_lens.max()
    return df[subtype_lens == max_subtype_strlen]


def sorted_subtype_ints(subtypes: pd.Series) -> List[List[int]]:
    """Get the list of subtypes as lists of integers sorted by subtype resolution

    Where the subtype resolution is determined by the number of periods ('.') in the subtype.

    Args:
        subtypes: pd.Series of subtype strings

    Return:
        list of subtypes as lists of integers sorted by subtype resolution
    """
    subtypes_ints = [[int(y) for y in x.split('.')] for x in subtypes.unique()]
    subtypes_ints.sort(key=lambda a: len(a))
    return subtypes_ints


def absent_downstream_subtypes(subtype: str, subtypes: pd.Series, scheme_subtypes: List[str]) -> Optional[List[str]]:
    """Find the downstream subtypes that are not present in the results

    Args:
        subtype: Final subtype result
        subtypes: Subtypes found
        scheme_subtypes: Possible subtyping scheme subtypes

    Returns:
         List of downstream subtypes that are not present in the results or `None` if all immediately downstream
         subtypes are present.
    """
    escaped_subtype = re.escape(subtype)
    re_subtype = re.compile(r'^{}\.\d+$'.format(escaped_subtype))
    downstream_subtypes = [s for s in scheme_subtypes if re_subtype.search(s)]
    absentees = [x for x in downstream_subtypes if not (subtypes == x).any()]
    return absentees if len(absentees) > 0 else None


def set_inconsistent_subtypes(st: Subtype, inconsistent_subtypes: List[str]) -> Subtype:
    """Save the list of inconsistent subtypes, if any, to the Subtype result

    Args:
        st: Subtype result
        inconsistent_subtypes: List of inconsistent subtypes

    Returns:

    """
    if len(inconsistent_subtypes) > 0:
        st.are_subtypes_consistent = False
        st.inconsistent_subtypes = inconsistent_subtypes
    else:
        st.are_subtypes_consistent = True
        st.inconsistent_subtypes = None
    return st


def empty_results(st: Subtype) -> pd.DataFrame:
    """Get a results DataFrame with 1 entry when no results are retrieved.

    When the output DataFrame for subtyping is empty, generate a "dummy" DataFrame with one row showing that subtyping
    was performed on the sample, but no results were generated because nothing was found.

    Args:
        st: Subtype results; should be a null result

    Returns:
        pd.DataFrame with 1 row with null result
    """
    return pd.DataFrame({0: dict(sample=st.sample,
                                 file_path=st.file_path,
                                 subtype=st.subtype,
                                 refposition=None,
                                 is_pos_tile=None,
                                 scheme=st.scheme,
                                 scheme_version=st.scheme_version,
                                 qc_status=st.qc_status,
                                 qc_message=st.qc_message)}).transpose()



# -*- coding: utf-8 -*-

from distutils.core import setup

from setuptools import find_packages

from bio_hansel import __version__, program_name, program_desc

classifiers = """
Development Status :: 3 - Alpha
Environment :: Console
License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)
Intended Audience :: Science/Research
Topic :: Scientific/Engineering
Topic :: Scientific/Engineering :: Bio-Informatics
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
Programming Language :: Python :: Implementation :: CPython
Operating System :: POSIX :: Linux
""".strip().split('\n')

setup(
    name=program_name,
    version=__version__,
    packages=find_packages(exclude=['tests']),
    url='https://github.com/phac-nml/{}'.format(program_name),
    license='Apache v2.0',
    author='Peter Kruczkiewicz',
    author_email='peter.kruczkiewicz@gmail.com',
    description=program_desc,
    keywords='Salmonella enterica Heidelberg Enteritidis SNP kmer subtyping Aho-Corasick',
    classifiers=classifiers,
    package_dir={program_name: program_name},
    package_data={program_name: ['data/*/*.fasta',]},
    install_requires=[
        'numpy>=1.12.1',
        'pandas>=0.20.1',
        'pyahocorasick>=1.1.6',
        'attrs',
    ],
    extras_require={
        'test': ['pytest>=3.0.7',],
    },
    entry_points={
        'console_scripts': [
            'hansel={}.main:main'.format(program_name),
        ],
    }
)



# -*- coding: utf-8 -*-

import logging
from typing import List, Any, Optional, Tuple, Union

import os
import re
from collections import defaultdict

from .const import SCHEME_FASTAS, REGEX_FASTQ, REGEX_FASTA
from .subtyping_params import SubtypingParams


def does_file_exist(filepath: str, force: bool):
    if filepath and os.path.exists(filepath):
        if not force:
            file_exists_err_fmt = 'File "{}" already exists! If you want to overwrite this output file run with opt "--force"'
            raise OSError(file_exists_err_fmt.format(filepath))
        else:
            logging.warning('File "{}" already exists, overwriting with "--force" - uh oh :S')


def genome_name_from_fasta_path(fasta_path: str) -> str:
    """Extract genome name from fasta filename

    Get the filename without directory and remove the file extension.

    Example:
        With fasta file path ``/path/to/genome_1.fasta``::

            fasta_path = '/path/to/genome_1.fasta'
            genome_name = genome_name_from_fasta_path(fasta_path)
            print(genome_name)
            # => "genome_1"

    Args:
        fasta_path (str): fasta file path

    Returns:
        str: genome name
    """
    filename = os.path.basename(fasta_path)
    filename = re.sub(r'\.gz$', '', filename)
    return re.sub(r'\.(fa|fas|fasta|fna|\w{1,})(\.gz)?$', '', filename)


def compare_subtypes(a: List[Any], b: List[Any]) -> bool:
    for x, y in zip(a, b):
        if x != y:
            return False
    return True


def find_inconsistent_subtypes(subtypes: List[List[int]]) -> List[str]:
    from collections import Counter
    incon = []
    for i in range(len(subtypes) - 1):
        a = subtypes[i]
        for j in range(i + 1, len(subtypes)):
            b = subtypes[j]
            is_consistent = compare_subtypes(a, b)
            if not is_consistent:
                incon.append((a, b))
    l = []
    for a, b in incon:
        astr = '.'.join([str(x) for x in a])
        bstr = '.'.join([str(x) for x in b])
        l += [astr, bstr]
    c = Counter(l)
    incon_subtypes = []
    for subtype, freq in c.most_common():
        if freq >= 1:
            incon_subtypes.append(subtype)
        else:
            break
    return incon_subtypes


def get_scheme_fasta(scheme: str) -> str:
    if scheme in SCHEME_FASTAS:
        scheme_fasta = SCHEME_FASTAS[scheme]['file']
    elif os.path.exists(scheme) and os.path.isfile(scheme):
        scheme_fasta = scheme
    else:
        raise FileNotFoundError('Could not find user-specified subtyping scheme fasta "%s"', scheme)
    return scheme_fasta


def get_scheme_params(scheme: str) -> Optional[SubtypingParams]:
    if scheme in SCHEME_FASTAS:
        return SCHEME_FASTAS[scheme]['subtyping_params']


def get_scheme_version(scheme: str) -> Optional[str]:
    if scheme in SCHEME_FASTAS:
        version = SCHEME_FASTAS[scheme]['version']  # type: str
        return version
    return None


def collect_fastq_from_dir(input_directory: str) -> List[Union[str, Tuple[List[str], str]]]:
    fastqs = []
    for x in os.listdir(input_directory):
        full_file_path = os.path.abspath(os.path.join(input_directory, x))
        if os.path.isfile(full_file_path) and REGEX_FASTQ.match(x):
            fastqs.append(full_file_path)
    if len(fastqs) > 0:
        logging.info('Found %s FASTQ files in %s',
                     len(fastqs),
                     input_directory)
        reads_from_dir = group_fastqs(fastqs)
        logging.info('Collected %s read sets from %s FASTQ files in %s',
                     len(reads_from_dir),
                     len(fastqs),
                     input_directory)
        return reads_from_dir
    return []


def group_fastqs(fastqs: List[str]) -> List[Tuple[List[str], str]]:
    """Group FASTQs based on common base filename

    For example, if you have 2 FASTQs:

    - reads_1.fastq
    - reads_2.fastq

    The common name would be `reads` and the files would be grouped based on that common name.

    Args:
        fastqs: FASTQ file paths

    Returns:
        list of grouped FASTQs grouped by common base filename
    """
    genome_fastqs = defaultdict(list)
    for fastq in fastqs:
        filename = os.path.basename(fastq)
        basefilename = re.sub(r'_\d', '', REGEX_FASTQ.sub(r'\1', filename))
        genome_fastqs[basefilename].append(fastq)
    return [(fastq_paths, genome_name) for genome_name, fastq_paths in genome_fastqs.items()]


def collect_fasta_from_dir(input_directory: str) -> List[Tuple[str, str]]:
    input_genomes = []
    for x in os.listdir(input_directory):
        full_file_path = os.path.abspath(os.path.join(input_directory, x))
        if os.path.isfile(full_file_path) and REGEX_FASTA.match(x):
            genome_name = genome_name_from_fasta_path(full_file_path)
            input_genomes.append((full_file_path, genome_name))
    return input_genomes


NT_SUB = {x: y for x, y in zip('acgtrymkswhbvdnxACGTRYMKSWHBVDNX', 'tgcayrkmswdvbhnxTGCAYRKMSWDVBHNX')}


def revcomp(s):
    """Reverse complement nucleotide sequence

    Args:
        s (str): nucleotide sequence

    Returns:
        str: reverse complement of `s` nucleotide sequence
    """
    return ''.join([NT_SUB[c] for c in s[::-1]])


def is_gzipped(p: str) -> bool:
    return bool(re.match(r'^.+\.gz$', p))


def init_subtyping_params(args: Optional[Any] = None,
                          scheme: Optional[str] = None) -> SubtypingParams:
    """Initialize subtyping parameters based on command-line arguments and scheme defaults

    Args:
        args: ArgumentParser.parse_args() output
        scheme: Scheme name e.g. "heidelberg"

    Returns:
        SubtypingParams with user-supplied values then scheme defaults then global defaults loaded
    """
    subtyping_params = get_scheme_params(scheme)
    if subtyping_params is None:
        subtyping_params = SubtypingParams()
    if args is not None:
        if args.low_cov_depth_freq:
            subtyping_params.low_coverage_depth_freq = args.low_cov_depth_freq
        if args.max_missing_tiles:
            subtyping_params.max_perc_missing_tiles = args.max_missing_tiles
        if args.min_ambiguous_tiles:
            subtyping_params.min_ambiguous_tiles = args.min_ambiguous_tiles
        if args.max_intermediate_tiles:
            subtyping_params.max_perc_intermediate_tiles = args.max_intermediate_tiles
        if args.low_cov_warning:
            subtyping_params.min_coverage_warning = args.low_cov_warning

    return subtyping_params
